import os

from PyQt5.QtCore import QDataStream, QIODevice, QObject
from PyQt5.QtNetwork import QTcpSocket


class TrayLauncherClient(QObject):
    def __init__(self, command, data):
        """Initiate the client.

        Args:
            command: str
            data: str
        """
        super().__init__()

        self.blockSize = 0

        self.client = QTcpSocket(self)

        self.command = command
        self.data = data

    @classmethod
    def check_connection(cls):
        """Attempts to connect to the tray launcher TCP server
        to check if there is already one tray launcher instance running
        """
        instance = cls("test", "empty")
        instance.client.connectToHost(
            "127.0.0.1", int(os.environ.get("TRAY_LAUNCHER_PORT", 7686)), QIODevice.ReadWrite
        )
        success_connect = instance.client.waitForConnected()

        if not success_connect:
            return False
        else:
            try:
                instance.client.write(bytes(instance.command + "\n", encoding="ascii"))
                instance.client.write(bytes((instance.data + "\n"), encoding="ascii"))
                instance.client.waitForDisconnected()
                return True
            except Exception:
                return False

    def attempt_connect(self):
        """Connects to the tray launcher server."""
        self.client.connectToHost(
            "127.0.0.1", int(os.environ.get("TRAY_LAUNCHER_PORT", 7686)), QIODevice.ReadWrite
        )
        success_connect = self.client.waitForConnected()

        if not success_connect:
            print(
                (
                    "Failed to connect to the launcher server. Run tray launcher first,"
                    " and check if the port supposedly used by the tray launcher is occupied."
                )
            )
            return

        self.client.readyRead.connect(self.read_from_server)

        self.client.write(bytes(self.command + "\n", encoding="ascii"))

        for entry in self.data:
            self.client.write(bytes((entry + "\n"), encoding="ascii"))

        self.client.waitForDisconnected()

    def read_from_server(self):
        instr = QDataStream(self.client)
        instr.setVersion(QDataStream.Qt_5_0)
        if self.blockSize == 0:
            if self.client.bytesAvailable() < 2:
                return
            self.blockSize = instr.readUInt16()
        if self.client.bytesAvailable() < self.blockSize:
            return
        print(str(instr.readString(), encoding="ascii"))



import argparse

from tray_launcher import cli, tray_launcher_client


def get_parser():
    parser = argparse.ArgumentParser(
        prog="tray_launcher", description="Manage scripts with the tray launcher"
    )

    launcher = parser.add_subparsers(dest="launcher")
    launcher.required = True

    p_start = launcher.add_parser("start", help="Starts scripts")
    p_start.add_argument(
        "scripts", type=str, nargs="*", metavar="script_stem", help="Scripts to be started"
    )

    launcher.add_parser("quit", help="Quits the tray launcher")

    p_terminate = launcher.add_parser("terminate", help="Terminates scripts")
    p_terminate.add_argument(
        "scripts",
        nargs="*",
        metavar="script_stem",
        type=str,
        help="Scripts to be terminated",
    )

    launcher.add_parser("run", help="Runs the tray launcher")

    p_load = launcher.add_parser("load", help="Loads scripts")
    p_load.add_argument(
        "scripts", nargs="*", metavar="script_path", type=str, help="Scripts to be loaded"
    )

    p_restart = launcher.add_parser("restart", help="Restarts scripts")
    p_restart.add_argument(
        "scripts",
        nargs="*",
        metavar="script_stem",
        type=str,
        help="Scripts to be restarted",
    )

    p_log = launcher.add_parser(
        "log",
        help=("Views scripts' logs."),
    )
    p_log.add_argument(
        "scripts",
        nargs="*",
        metavar="script_stem",
        type=str,
        help="Scripts whose log is to be viewed",
    )
    p_log.add_argument("-a", "--all", action="store_true", help="View all logs")

    p_focus = launcher.add_parser("focus", help="Focus scripts")
    p_focus.add_argument(
        "scripts", nargs="*", metavar="script_stem", type=str, help="Scripts to be focused"
    )

    p_list = launcher.add_parser("list", help="Lists scripts")
    p_list.add_argument("-r", "--running", action="store_true", help="View all running scripts")
    p_list.add_argument("-a", "--all", action="store_true", help="View all scripts")

    return parser


def get_print_pre_command(launcher, scripts):
    if launcher == "start":
        print_pre_command = "Starting {}.".format(scripts)
    elif launcher == "terminate":
        print_pre_command = "Terminating {}.".format(scripts)
    elif launcher == "restart":
        print_pre_command = "Restarting {}.".format(scripts)
    elif launcher == "load":
        print_pre_command = "Loading {}.".format(scripts)
    elif launcher == "focus":
        print_pre_command = "Bringing {} to the top.".format(scripts)
    return print_pre_command


def dispatch_command(args):
    if args.launcher == "run":
        cli.run_pythonw()
        return

    print_pre_command = ""
    if args.launcher in ["start", "terminate", "restart", "load", "focus"]:
        print_pre_command = get_print_pre_command(args.launcher, args.scripts)
        commands = (args.launcher, args.scripts)
    elif args.launcher == "quit":
        print_pre_command = "Quitting tray launcher."
        commands = ("quit", [])
    elif args.launcher == "log":
        if args.all:
            print_pre_command = "Showing all logs."
            commands = ("all_logs", [])
        else:
            print_pre_command = "Showing logs for {}.".format(args.scripts)
            commands = ("log", args.scripts)
    elif args.launcher == "list":
        if args.running:
            print_pre_command = "Running scripts: "
            commands = ("list_current", [])
        elif args.all:
            print_pre_command = "All available scripts: "
            commands = ("list", [])
        else:
            print(
                (
                    "tray_launcher list: error: at least one of the following"
                    " arguments are required: -a/--all, -r/--running"
                )
            )
            return
    else:
        print("tray_launcher: error: {} is an unrecognized command".format(args.launcher))
        return

    print(print_pre_command)
    tray_launcher_client.TrayLauncherClient(*commands).attempt_connect()


def main():
    args = get_parser().parse_args()
    dispatch_command(args)


if __name__ == "__main__":
    main()



import logging
import os
import signal
import subprocess
import time as _t
from pathlib import Path


class ChildScript:
    ENCODING = "utf-8"

    def __init__(self, script_path_str, logging_log):
        self.script_path_str = script_path_str
        self.script_path = Path(self.script_path_str)
        self.child_script = None
        self.child_script_PID = -1
        self.current_PIDs = []

        logging.basicConfig(
            filename=logging_log,
            level=logging.INFO,
            format="%(asctime)s %(message)s",
        )

    def start_script(self):
        t = _t.localtime(_t.time())

        log_file = self.get_log_path()

        try:
            log_directory = log_file / (
                str(t.tm_year) + "_" + str(t.tm_mon).zfill(2) + "_" + str(t.tm_mday).zfill(2)
            )
            log_directory.mkdir(parents=True, exist_ok=True)
        except Exception as err:
            logging.error(err + ": Failed to create new directory for ChildScript outputs.")
            raise

        self.log_path = log_directory / "{}-{}_{}_{}.log".format(
            self.script_path.stem,
            str(t.tm_hour).zfill(2),
            str(t.tm_min).zfill(2),
            str(t.tm_sec).zfill(2),
        )

        try:
            self.outputs_file = open(self.log_path, "a")
        except Exception as err:
            logging.error(err + ": Failed to open/create a file for ChildScript outputs.")
            raise

        self.child_script = subprocess.Popen(
            '"' + self.script_path_str + '"',
            encoding=self.ENCODING,
            stdout=self.outputs_file,
            stderr=self.outputs_file,
            creationflags=subprocess.CREATE_NO_WINDOW,
        )

        self.child_script_PID = self.child_script.pid

    def terminate_script(self):
        self.update_current_PIDs()
        try:
            for pid in self.current_PIDs:
                subprocess.run(["taskkill", "/F", "/T", "/PID", str(pid)])
        except Exception:
            logging.error(
                "Child Script PID: "
                + str(self.child_script_PID)
                + ": Error when terminating child processes."
            )

        try:
            self.outputs_file.close()
            os.kill(self.child_script_PID, signal.SIGTERM)
        except OSError:
            logging.error(
                "Child Script PID: "
                + str(self.child_script_PID)
                + ": The Popen process is not running."
            )
            return

    def get_log_path(self):
        user_home = Path.home() / ".tray_launcher"
        log_file = user_home / "logs"
        return log_file

    def is_active(self):
        """Checks if there is any subprocess still running."""
        self.update_current_PIDs()

        if self.child_script.poll() is None:
            return True
        elif self.current_PIDs:
            return True
        return False

    def update_current_PIDs(self):
        self.current_PIDs = []
        wmic_ = subprocess.run(
            "wmic process where (ParentProcessId={}) get ProcessId".format(
                str(self.child_script_PID)
            ),
            encoding=self.ENCODING,
            stdout=subprocess.PIPE,
            shell=True,
        )
        for line in wmic_.stdout.split("\n"):
            if line != "" and line.split(" ")[0] != "ProcessId":
                self.current_PIDs.append(int(line))



from PyQt5.QtCore import QObject
from win32con import (
    HWND_NOTOPMOST,
    HWND_TOPMOST,
    SW_RESTORE,
    SWP_NOMOVE,
    SWP_NOSIZE,
    SWP_SHOWWINDOW,
)
from win32gui import EnumWindows, IsWindowEnabled, IsWindowVisible, SetWindowPos, ShowWindow
from win32process import GetWindowThreadProcessId

from tray_launcher import child_script


class ChildScriptManager(QObject):
    def __init__(self):
        super().__init__()

        # key: timestamp
        # value: ChildScript
        self.running_child_scripts = {}

    def run_new(self, args):
        """Starts a new script.

        Args:
            args: ((Path, int), str), the path to the script,
                the timestamp of the ChildScript.
                Plus the path to the log file for the tray launcher.
        """
        c = child_script.ChildScript(str(args[0][0]), args[1])
        c.start_script()
        self.running_child_scripts[args[0][1]] = c

    def show(self, args):
        """Brings windows associated with a script to the foreground.

        Args:
            args: ((Path, int), str), the path to the script,
                the timestamp of the ChildScript.
                Plus the path to the log file for the tray launcher.
        """
        self.running_child_scripts[args[0][1]].update_current_PIDs()
        self.bring_to_front(self.running_child_scripts[args[0][1]].current_PIDs)

    def terminate(self, timestamp):
        """Terminates the script started at the time specified
            by the argument timestamp. Plus the path to the log file for the tray launcher.

        Args:
            args: (int, str), the timestamp of the ChildScript.
        """
        self.running_child_scripts[timestamp[0]].terminate_script()
        del self.running_child_scripts[timestamp[0]]

    def get_hwnds_for_PID(self, pid):
        """Get WINDOW handles for windows associated with the given PID.

        Args:
            pid: int, the PID whose window handles are to be found.
        """

        def callback(hwnd, hwnds):
            if IsWindowVisible(hwnd) and IsWindowEnabled(hwnd):
                _, found_pid = GetWindowThreadProcessId(hwnd)
                if found_pid == pid:
                    hwnds.append(hwnd)
            return True

        hwnds = []
        EnumWindows(callback, hwnds)
        return hwnds

    def bring_to_front(self, pids):
        """Brings the windows of processes with the given pids to the foreground.

        Arg:
            pids: list of int, PIDs of the processes whose windows are
                to be brought to the foreground.
        """
        for pid in pids:
            for window_handle in self.get_hwnds_for_PID(pid):
                ShowWindow(window_handle, SW_RESTORE)
                SetWindowPos(window_handle, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE + SWP_NOSIZE)
                SetWindowPos(window_handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE + SWP_NOSIZE)
                SetWindowPos(
                    window_handle,
                    HWND_NOTOPMOST,
                    0,
                    0,
                    0,
                    0,
                    SWP_SHOWWINDOW + SWP_NOMOVE + SWP_NOSIZE,
                )



import logging
import os
import shutil as _su
import time as _t
from functools import partial
from pathlib import Path

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (
    QAction,
    QDesktopWidget,
    QFileDialog,
    QLabel,
    QMainWindow,
    QMenu,
    QMessageBox,
    QSystemTrayIcon,
    QVBoxLayout,
    QWidget,
    qApp,
)

from tray_launcher import child_script, child_script_manager


class TrayLauncherGUI(QMainWindow):
    def __init__(self):
        home_path = Path(child_script.__file__).parent
        user_path = Path.home()

        user_home = user_path / ".tray_launcher"

        self.LOGS = user_home / "logs"
        self.AVAILABLE_SCRIPTS = user_home / "scripts"

        self.icon = str(home_path / "icons" / "tray_icon.png")
        self.check_mark = str(home_path / "icons" / "check_mark.png")

        super().__init__()
        self.script_manager = child_script_manager.ChildScriptManager()
        t = _t.localtime(_t.time())
        try:
            self._log_directory = self.LOGS / (
                str(t.tm_year) + "_" + str(t.tm_mon).zfill(2) + "_" + str(t.tm_mday).zfill(2)
            )
            self._log_directory.mkdir(parents=True, exist_ok=True)
        except Exception:
            # print(err + ": Failed to create new directory for outputs")
            raise

        self.tray_launcher_log = self._log_directory / "tray_launcher.log"
        logging.basicConfig(
            filename=self.tray_launcher_log,
            level=logging.INFO,
            format="%(asctime)s %(message)s",
        )

        try:
            self.AVAILABLE_SCRIPTS.mkdir(parents=True, exist_ok=True)
        except Exception as err:
            logging.error(err + ": Failed to create new directory for available scripts")
            raise

        logging.info("Tray Launcher Started.")

        self.script_count = 0

        # key: string:                          script.stem
        # value: tuple (int, QMenu):            (timestamp, three_menu)
        self.currently_running_scripts = {}

        # key: script.stem
        # value: QAction
        self.available_scripts = {}

        self.init_ui()

    def init_ui(self):
        self.trayicon = QSystemTrayIcon(self)
        self.trayicon.setIcon(QIcon(self.icon))
        self.trayicon.setVisible(True)

        self.context_menu = QMenu(self)
        load_new_script = QAction("Load New Script(s)", self)
        load_new_script.triggered.connect(
            partial(self.load_scripts_from_file_dialogue, Path.home())
        )

        self.view_all = QMenu("Start a Script", self)
        self.add_available_scripts(self.view_all)

        self.view_in_directory = self.view_all.addAction("[View in Directory]")
        self.view_in_directory.triggered.connect(
            partial(self.open_script_from_file_dialogue, self.AVAILABLE_SCRIPTS)
        )

        self.currently_running_section = self.context_menu.addSection("Currently Running")

        self.none_currently_running = self.context_menu.addAction("None")
        self.none_currently_running.setEnabled(False)

        self.bottom_separator = self.context_menu.addSeparator()

        self.context_menu.addMenu(self.view_all)

        self.context_menu.addAction(load_new_script)

        logs = self.context_menu.addAction("All Logs")
        logs.triggered.connect(partial(self.show_logs, self.LOGS))

        help_ = self.context_menu.addAction("Help")
        help_.triggered.connect(self.show_help)

        self.quit_ = self.context_menu.addAction("Quit")
        self.quit_.triggered.connect(self.quit)

        self.context_menu.aboutToShow.connect(self.prepare_context_menu)
        self.trayicon.setContextMenu(self.context_menu)

        self.check_available_scripts()
        self.show()
        super().resize(0, 0)
        self.trayicon.show()

    def to_loaded_path(self, path_given):
        """Returns a Path that resembles one pointing to the "scripts" directory.
            Changes the parent of the path_given to the "scripts" directory and modifies its
            suffix to .bat

        Args:
            path_given: Path
        """
        try:
            loaded_path = Path(self.AVAILABLE_SCRIPTS / path_given).with_suffix(".bat")
        except ValueError:
            return None
        return loaded_path

    def start_new_script(self, script_path):
        """Starts a new script.

        Args:
            script_path: Path, path to the script to be started.
        """
        self.prepare_context_menu()

        if script_path.stem in self.currently_running_scripts:
            return

        timestamp = _t.time()
        args = (script_path, timestamp)
        three_menu = QMenu(script_path.stem, self)

        showAction = QAction("Show", self)
        showAction.triggered.connect(partial(self.show_script, (args, three_menu)))
        three_menu.addAction(showAction)

        restartAction = QAction("Restart", self)
        restartAction.triggered.connect(partial(self.restart_script, (args, three_menu)))
        three_menu.addAction(restartAction)

        terminateAction = QAction("Terminate", self)
        terminateAction.triggered.connect(partial(self.terminate_script, (args, three_menu)))
        three_menu.addAction(terminateAction)

        self.run_in_manager(args, self.script_manager.run_new)

        self.none_currently_running.setVisible(False)
        self.script_count += 1

        logging.info("{} was started.".format(script_path.stem))

        logAction = QAction("Log", self)
        logAction.triggered.connect(
            partial(
                self.show_logs,
                self.script_manager.running_child_scripts[timestamp].log_path,
            )
        )
        three_menu.insertAction(restartAction, logAction)

        self.context_menu.insertMenu(self.bottom_separator, three_menu)

        self.currently_running_scripts[script_path.stem] = (timestamp, three_menu)

        three_menu.menuAction().setIcon(QIcon(self.check_mark))
        self.available_scripts[script_path.stem].setEnabled(False)

    def show_script(self, args):
        """Bring windows associated with a script to the foreground, if any.

        Args:
            args: ((Path, int), QMenu), the path to the script, the timestamp of the ChildScript,
                and the QMenu associated with this script.
        """
        self.run_in_manager(args[0], self.script_manager.show)

        logging.info(
            "{} was brought to the front.".format(args[0][0].stem)
            + " Processes with PIDs {} are running.".format(
                self.script_manager.running_child_scripts[args[0][1]].current_PIDs
            )
        )

    def terminate_script(self, args):
        """Terminate a script.

        Args:
            args: ((Path, int), QMenu), the path to the script,
                the timestamp of the ChildScript, and the QMenu associated with this script.
        """
        if args[0][0].stem in self.currently_running_scripts:
            self.run_in_manager(args[0][1], self.script_manager.terminate)
            self.context_menu.removeAction(args[1].menuAction())

            del self.currently_running_scripts[args[0][0].stem]

            self.script_count -= 1

            if self.script_count == 0:
                self.none_currently_running.setVisible(True)

            if (args[0][0]).stem in self.available_scripts:
                self.available_scripts[(args[0][0]).stem].setIcon(QIcon())
                self.available_scripts[(args[0][0]).stem].setEnabled(True)

            logging.info("{} was terminated through Tray Launcher.".format(args[0][0].stem))

    def restart_script(self, args):
        """Restarts a script.

        Args:
            args: ((Path, int), QMenu), the path to the script, the timestamp of the ChildScript,
                and the QMenu associated with this script.
        """
        self.terminate_script(args)
        self.start_new_script(args[0][0])

        logging.info("{} was restarted.".format(args[0][0].stem))

    def run_in_manager(self, args, func):
        """Runs a function through the script manager.

        Args:
            args: (str, int), the path to the script, the timestamp of the ChildScript.
                in case of func=terminate_script, args: int
            func: an instance function of ChildScriptManager.
        """
        args = (args, self.tray_launcher_log)
        dummy_action = QAction(self)
        dummy_action.triggered.connect(partial(func, args))
        dummy_action.trigger()

    def add_available_scripts(self, target_menu):
        """Loads all .bat files in the \\scripts directory to the menu specified.

        Args:
            target_menu: QMenu, the menu to be loaded with .bat file stems.
        """
        self.available_scripts.clear()
        target_menu.clear()

        for file_path in Path.iterdir(self.AVAILABLE_SCRIPTS):
            if file_path.is_file() and file_path.suffix == ".bat":
                st = file_path.stem

                action = target_menu.addAction(st)
                action.triggered.connect(partial(self.start_new_script, file_path))
                self.available_scripts[st] = action

                if st in self.currently_running_scripts:
                    self.available_scripts[st].setIcon(QIcon(self.check_mark))
                    self.available_scripts[st].setEnabled(False)
            else:
                if file_path.is_file:
                    file_path.unlink()

    def check_available_scripts(self):
        """Reloads .bat files in the \\scripts directory to the view_all menu"""
        self.add_available_scripts(self.view_all)

        self.view_in_directory = QAction("[View in Directory]", self)
        self.view_in_directory.triggered.connect(
            partial(self.open_script_from_file_dialogue, self.AVAILABLE_SCRIPTS)
        )
        self.view_all.addAction(self.view_in_directory)

    def prepare_context_menu(self):
        """Checks if scripts are still running; if not, remove them from the menu.
        Also rebuilds the the view_all menu
        """
        to_del = []
        for (
            timestamp,
            child_script_obj,
        ) in self.script_manager.running_child_scripts.items():
            if not child_script_obj.is_active():
                child_script_obj.outputs_file.close()

                to_del.append(timestamp)
                self.context_menu.removeAction(
                    self.currently_running_scripts[child_script_obj.script_path.stem][
                        1
                    ].menuAction()
                )
                del self.currently_running_scripts[child_script_obj.script_path.stem]

                self.script_count -= 1
                if self.script_count == 0:
                    self.none_currently_running.setVisible(True)
                logging.info(
                    "{} has already been terminated. Now removed from the menu.".format(
                        child_script_obj.script_path_str
                    )
                )

        for ts in to_del:
            del self.script_manager.running_child_scripts[ts]

        self.check_available_scripts()

    def load_scripts_from_file_dialogue(self, dir):
        """
        Args:
            dir: Path, path of the directory from which a file is to be loaded.
        """
        fnames, _ = QFileDialog.getOpenFileNames(
            self, "Loading New Script(s) (.bat only)", str(dir), ("Images (*.bat)")
        )

        for file_name in fnames:
            self.load_script(Path(file_name))

    def open_script_from_file_dialogue(self, file_dialogue_path):
        """Starts a file selected from the directory specified
            by the argument file_dialogue_path.
            If the file is not in the \\scripts directory, load it.

        Args:
            file_dialogue_path: Path, path of the directory from which a file is to be loaded.
        """
        file_name, _ = QFileDialog.getOpenFileName(
            self, "Starting a Script (.bat only)", str(file_dialogue_path), ("Images (*.bat)")
        )

        if file_name != "":
            file_path = Path(file_name)

            self.run_new_file(file_path)

    def run_new_file(self, file_path):
        """Attempts to run the given argument as a .bat script

        Args:
            file_path: Path
        """

        if not (file_path.is_file()):
            logging.info("Only .bat file is accepted.")
            return False

        elif file_path.parent == self.AVAILABLE_SCRIPTS:
            self.start_new_script(file_path)
            return True
        else:
            if self.load_script(file_path):
                file_path = self.to_loaded_path(file_path.stem)
                self.start_new_script(file_path)
            return True

    def load_script(self, script_path):
        """Loads the specified file to the \\scripts directory.
            If there is a file with the same name, asks the user
            if they wish to replace.

        Args:
            script_path: Path, the path to the file to be loaded.
        """
        isDuplicateName = False

        if script_path.is_file() and script_path.suffix == ".bat":
            for existing_file_path in Path.iterdir(self.AVAILABLE_SCRIPTS):
                if existing_file_path.stem == script_path.stem:
                    isDuplicateName = True
                    break

            if not isDuplicateName:
                _su.copy(script_path, self.AVAILABLE_SCRIPTS)

                logging.info("{} was loaded to \\scripts.".format(str(script_path)))
                self.prepare_context_menu()
                return True
            else:
                self.resize(1, 1)
                self.showMinimized()
                self.showMaximized()
                self.resize(0, 0)
                b = QMessageBox()
                b.setWindowFlag(Qt.WindowStaysOnTopHint)
                replace_reply = b.question(
                    self,
                    "Replace File",
                    "A file named {} already exists in "
                    "\\scripts. Do you want to replace it?".format(script_path.name),
                    QMessageBox.Yes | QMessageBox.No,
                    QMessageBox.No,
                )
                if replace_reply == QMessageBox.Yes:
                    try:
                        _su.copy(script_path, self.AVAILABLE_SCRIPTS)
                        logging.info("{} was replaced in \\scripts.".format(str(script_path)))
                        self.prepare_context_menu()
                        return True
                    except _su.SameFileError:
                        logging.info("Same file was uploaded!")
                        return True
                else:
                    return False
        else:
            return False

    def quit(self):
        self.resize(1, 1)
        self.showMinimized()
        self.showMaximized()
        self.resize(0, 0)
        b = QMessageBox()
        b.setWindowFlag(Qt.WindowStaysOnTopHint)
        replace_reply = b.question(
            self,
            "Quit Tray Launcher",
            "Do you want to quit tray launcher?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.Yes,
        )
        if replace_reply == QMessageBox.Yes:
            for tuple in self.currently_running_scripts.values():
                self.script_manager.terminate(tuple[0])
            logging.info("Tray Launcher Exited.")
            self.script_manager.deleteLater()
            qApp.quit()

    def show_logs(self, log_path):
        """Displays the file specified.

        Args:
            log_path: Path, the path to the file to be opened.
        """
        logging.info("Logs {} were opened.".format(log_path))
        os.startfile(log_path)

    def show_help(self):
        """Displays a Help window in the middle of the screen"""
        self.help_window = QWidget()
        self.help_window.resize(100, 100)

        Rect = self.help_window.frameGeometry()
        center = QDesktopWidget().availableGeometry().center()
        Rect.moveCenter(center)
        self.help_window.move(Rect.topLeft())

        self.help_window.setWindowTitle("Help")

        layout = QVBoxLayout()
        layout.addWidget(
            QLabel(
                '1. Load scripts by clicking the "Load New Scripts(s)" option \nand'
                + " select scripts you wish to run in the future. \n2. Move over"
                + ' "Start a Script" to select one script to run. \n 3. To use CLI,'
                + ' run "launcher -h" to view all allowed commands.'
            )
        )
        self.help_window.setLayout(layout)
        self.help_window.show()



import logging
import os
import subprocess
import sys
import time as _t
from functools import partial
from pathlib import Path

from PyQt5.QtCore import QByteArray, QDataStream, QIODevice, QObject
from PyQt5.QtGui import QIcon
from PyQt5.QtNetwork import QHostAddress, QTcpServer
from PyQt5.QtWidgets import QApplication, qApp

from tray_launcher import gui, tray_launcher_client


class TrayLauncherCLI(QObject):
    def __init__(self):
        super().__init__()
        self.message_to_client = []

        self.server = QTcpServer(self)
        port = int(os.environ.get("TRAY_LAUNCHER_PORT", 7686))
        address = QHostAddress("127.0.0.1")
        if not self.server.listen(address, port):
            logging.warning(
                "Failed to listen to port: "
                + str(port)
                + ". See README.md to switch to an available port."
            )
            return
        self.server.newConnection.connect(self.connection)

        self.gui = gui.TrayLauncherGUI()

    def connection(self):
        """Processes the list passed from the client, and writes response back."""
        client_connection = self.server.nextPendingConnection()

        client_connection.waitForReadyRead()

        data = []
        while not client_connection.atEnd():
            data.append(str(client_connection.readLine(), encoding="ascii")[0:-1])

        dispatchers = {
            "test": self.test,
            "start": self.start,
            "terminate": self.terminate,
            "list": self.list_all,
            "list_current": self.list_current,
            "load": self.load,
            "restart": self.restart,
            "log": self.log,
            "all_logs": self.all_logs,
            "focus": self.focus,
            "quit": self.quit,
        }

        try:
            dispatcher = dispatchers[data[0]]
            dispatcher(data)
        except KeyError:
            self.process_invalid_command(data)

        self.write_to_client(client_connection)
        client_connection.disconnected.connect(client_connection.deleteLater)
        client_connection.disconnectFromHost()

    def test(self, data):
        """Processes the "test" command."""
        self.message_to_client.append(" ")

    def start(self, data):
        """Process the "start" command. Start new scripts."""
        for path_str in data[1:]:
            file_path = self.gui.to_loaded_path(Path(path_str))
            if file_path is not None:
                if file_path.stem in self.gui.currently_running_scripts:
                    self.message_to_client.append(
                        "Cannot run "
                        + "a script with the same stem as one of the currently running scripts."
                    )
                    return
                if self.gui.run_new_file(file_path):
                    self.message_to_client.append("SUCCESS: {} is now running.".format(path_str))
                else:
                    self.message_to_client.append(
                        "{} is not valid. Only .bat files are accepted.".format(path_str)
                    )
            else:
                self.message_to_client.append("{} is not valid.".format(path_str))

    def terminate(self, data):
        """Process the "terminate" command. Terminate scripts that are running."""
        for path_str in data[1:]:
            self._on_running_script(self.gui.terminate_script, path_str, " is terminated.")

    def list_all(self, data):
        """Processes the "list -a" command. Writes loaded scripts' stems to the string
        which will be written back to the client.
        """
        self.gui.available_scripts.clear()
        self.gui.view_all.clear()

        for file_path in Path.iterdir(self.gui.AVAILABLE_SCRIPTS):
            if file_path.is_file() and file_path.suffix == ".bat":
                st = file_path.stem

                action = self.gui.view_all.addAction(st)
                action.triggered.connect(partial(self.gui.start_new_script, file_path))
                self.gui.available_scripts[st] = action

                if st in self.gui.currently_running_scripts:
                    self.gui.available_scripts[st].setIcon(QIcon(self.gui.check_mark))
                    self.gui.available_scripts[st].setEnabled(False)

                    self.message_to_client.append(
                        "{} \t \t \t \t \t \t(currently running)".format(st)
                    )
                else:
                    self.message_to_client.append("{}".format(st))
            else:
                if file_path.is_file:
                    file_path.unlink()

    def list_current(self, data):
        """Processes the "list -r" command. Writes currently running scripts' stems to the string
        which will be written back to the client.
        """
        for st in self.gui.currently_running_scripts:
            self.message_to_client.append("{}".format(st))

    def load(self, data):
        """Processes the "load" command. Loads scripts to the 'scripts' directory"""
        for path_str in data[1:]:
            file_path = Path(path_str)
            if not (self.gui.load_script(file_path)):
                self.message_to_client.append(
                    "{} is not loaded. (Only .bat file is accepted)".format(path_str)
                )
            else:
                self.message_to_client.append("SUCCESS: {} is loaded.".format(path_str))

    def restart(self, data):
        """Processes the "restart" command. Restarts scripts."""
        for path_str in data[1:]:
            self._on_running_script(self.gui.restart_script, path_str, " is restarted.")

    def log(self, data):
        """Processes the "log" command. Brings up the log file of the scripts."""
        for path_str in data[1:]:
            self._on_running_script(self.gui.show_logs, path_str, "")

    def all_logs(self, data):
        self.gui.show_logs(self.gui.LOGS)
        self.message_to_client.append("SUCCESS: all logs are shown.")

    def focus(self, data):
        """Processes the "focus" command. Brings the scripts to the foreground."""
        for path_str in data[1:]:
            self._on_running_script(self.gui.show_script, path_str, " is brought to the front.")

    def quit(self, data):
        """Processes the "quit" command. Quits the tray launcher without prompting."""
        for tuple in self.gui.currently_running_scripts.values():
            self.gui.script_manager.terminate(tuple[0])
        logging.info("Tray Launcher Exited.")
        self.gui.script_manager.deleteLater()
        qApp.quit()

    def process_invalid_command(self, data):
        """Processes an invalid command."""
        self.message_to_client.append("{} is an invalid command.".format(data[0]))

    def _on_running_script(self, func, path_str, success_message):
        file_path = self.gui.to_loaded_path(Path(path_str))
        if file_path is not None:
            if (
                file_path.is_file()
                and file_path.parent == self.gui.AVAILABLE_SCRIPTS
                and file_path.stem in self.gui.currently_running_scripts
            ):
                if func == self.gui.show_logs:
                    self.gui.show_logs(
                        self.gui.script_manager.running_child_scripts[
                            self.gui.currently_running_scripts[file_path.stem][0]
                        ].log_path
                    )
                    self.message_to_client.append(
                        "SUCCESS: log of {} is shown.".format(file_path.stem)
                    )
                else:
                    func(
                        (
                            (file_path, self.gui.currently_running_scripts[file_path.stem][0]),
                            self.gui.currently_running_scripts[file_path.stem][1],
                        )
                    )
                    self.message_to_client.append(
                        "SUCCESS: {}".format(file_path.stem) + success_message
                    )
            elif file_path.is_file() and file_path.parent == self.gui.AVAILABLE_SCRIPTS:
                self.message_to_client.append("{} is not running.".format(path_str))
            else:
                self.message_to_client.append("{} is not valid.".format(path_str))
        else:
            self.message_to_client.append("{} is not valid.".format(path_str))

    def write_to_client(self, connection):
        """Passes the string (self.message_to_client) created by the server to the client."""

        if not self.message_to_client:
            return

        message = "\n".join(self.message_to_client)
        self.message_to_client = []

        block = QByteArray()
        out = QDataStream(block, QIODevice.ReadWrite)
        out.setVersion(QDataStream.Qt_5_0)
        out.writeUInt16(0)
        message = bytes(message, encoding="ascii")
        out.writeString(message)
        out.device().seek(0)
        out.writeUInt16(block.size() - 2)
        connection.write(block)


def main():
    app = QApplication(sys.argv)
    app.setStyle("Fusion")
    lc = TrayLauncherCLI()
    sys.exit(app.exec_())
    lc.deleteLater()


def run_pythonw():
    instance_already_exists = tray_launcher_client.TrayLauncherClient.check_connection()
    if instance_already_exists:
        print("There is already an instance of tray launcher running. Terminating now.")
        return

    HOME_PATH = Path(__file__).parent / "cli.py"

    t = _t.localtime(_t.time())

    log_directory = (
        Path.home()
        / ".tray_launcher"
        / "logs"
        / (str(t.tm_year) + "_" + str(t.tm_mon).zfill(2) + "_" + str(t.tm_mday).zfill(2))
    )

    try:
        log_directory.mkdir(parents=True, exist_ok=True)
    except Exception as err:
        print(err + ": Failed to create new directory for tray launcher outputs")
        raise
    with open(log_directory / "tray_launcher.log", "a") as launcher_log:
        subprocess.Popen(
            "python " + '"' + str(HOME_PATH) + '"',
            encoding="utf-8",
            creationflags=subprocess.CREATE_NO_WINDOW,
            stdout=launcher_log,
            stderr=launcher_log,
        )

    print("Tray Launcher is running...")


if __name__ == "__main__":
    main()
